/* WARNING: THIS FILE IS AUTOMATICALLY GENERATED FROM A .SM FILE.
 * Changes made here will certainly be overwritten.
 */

/* 
 * (C) 2003 Clemson University and The University of Chicago 
 *
 * See COPYING in top-level directory.
 */

/* pvfs2_client_getattr_pcache_sm
 *
 * The purpose of this state machine is to resolve a segment (one
 * component of a path) and a parent handle down to an object handle.
 * This object handle might already be present in a local cache (the
 * ncache) or could be obtained from a remote server.
 *
 * This state machine uses the msgpair structure in the PINT_client_sm
 * structure in order to perform the message passing.  It assumes that
 * this has been previously set up by the state machine above this
 * one.  In particular the following fields of the msgpair structure
 * must be filled in:
 *
 * - req (unencoded lookup request)
 * - fs_id
 * - handle
 * - svr_addr
 * - comp_fn - called to pull attributes out of decoded response and store
 *   them in the appropriate place for the state machine jumping to this one
 *   (e.g. copies object handle out in the remove case)
 *
 * This state machine in turns jumps to the pvfs2_msgpairarray_sm to
 * perform messaging, if necessary.
 *
 * NOTE: the comp_fn can cause a non-zero error_code to be passed back
 * up from the msgpair state machine and through this state machine
 * back to the one that jumped to here.  so be aware!
 *
 * TODO: sometimes we will get new attributes back with the lookup
 * response.  we should add these into the pcache as appropriate.
 *
 * TODO: This state machine does not handle resolving multiple segments.
 * It simply tries to resolve the first segment of the path in the msgpair or
 * sends the whole path to the msgpairarray_sm. This sm relies on sys-lookup 
 * to handle its responses and server responses and keep up with how much of 
 * the path has been resolved. That funcitonality might be better placed here 
 * since all sys-lookup cares about is the final resolved handle.
 */

#include <string.h>
#include <assert.h>

#include "client-state-machine.h"
#include "pvfs2-debug.h"
#include "job.h"
#include "gossip.h"
#include "str-utils.h"

#include "ncache.h"
#include "PINT-reqproto-encode.h"
#include "pvfs2-internal.h"

enum
{
    LOOKUP_NCACHE_MISS = 1
};


static PINT_sm_action lookup_ncache_lookup(
	struct PINT_smcb *smcb, job_status_s *js_p);

static struct PINT_state_s ST_lookup;
static struct PINT_pjmp_tbl_s ST_lookup_pjtbl[];
static struct PINT_tran_tbl_s ST_lookup_trtbl[];
static struct PINT_state_s ST_xfer_msgpair;
static struct PINT_pjmp_tbl_s ST_xfer_msgpair_pjtbl[];
static struct PINT_tran_tbl_s ST_xfer_msgpair_trtbl[];

struct PINT_state_machine_s pvfs2_client_lookup_ncache_sm = {
	.name = "pvfs2_client_lookup_ncache_sm",
	.first_state = &ST_lookup
};

static struct PINT_state_s ST_lookup = {
	 .state_name = "lookup" ,
	 .parent_machine = &pvfs2_client_lookup_ncache_sm ,
	 .flag = SM_RUN ,
	 .action.func = lookup_ncache_lookup ,
	 .pjtbl = NULL ,
	 .trtbl = ST_lookup_trtbl 
};

static struct PINT_tran_tbl_s ST_lookup_trtbl[] = {
	{ .return_value = LOOKUP_NCACHE_MISS ,
	 .next_state = &ST_xfer_msgpair },
	{ .return_value = -1 ,
	 .flag = SM_RETURN }
};

static struct PINT_state_s ST_xfer_msgpair = {
	 .state_name = "xfer_msgpair" ,
	 .parent_machine = &pvfs2_client_lookup_ncache_sm ,
	 .flag = SM_JUMP ,
	 .action.nested = &pvfs2_msgpairarray_sm ,
	 .pjtbl = NULL ,
	 .trtbl = ST_xfer_msgpair_trtbl 
};

static struct PINT_tran_tbl_s ST_xfer_msgpair_trtbl[] = {
	{ .return_value = -1 ,
	 .flag = SM_RETURN }
};

# 82 "src/client/sysint/lookup-ncache.sm"


static int lookup_ncache_lookup(struct PINT_smcb *smcb,
				job_status_s *js_p)
{
    struct PINT_client_sm *sm_p = PINT_sm_frame(smcb, PINT_FRAME_CURRENT);
    int ret = -PVFS_EINVAL;
    PVFS_object_ref parent_ref, object_ref;
    char segment[PVFS_NAME_MAX + 1] = {0};
    char* temp;
    
    gossip_debug(GOSSIP_CLIENT_DEBUG, "lookup_ncache state: lookup\n");

    assert(js_p->error_code == 0);

    parent_ref.fs_id = sm_p->msgpair.fs_id;
    parent_ref.handle = sm_p->msgpair.handle;

    assert(parent_ref.handle != PVFS_HANDLE_NULL);
    assert(parent_ref.fs_id != PVFS_FS_ID_NULL);
    
    /* Stripping out first segment for ncache lookup */
    strncpy(segment,sm_p->msgpair.req.u.lookup_path.path,PVFS_NAME_MAX);

    assert(segment[0] != '/');
    temp = index(segment,'/');
    if(temp != NULL)
    {
        temp[0] = '\0';
    }

    gossip_debug(GOSSIP_CLIENT_DEBUG, 
                 "lookup_ncache state: "
                 "lookup_ncache_lookup on segment [%s]\n",
                 segment);

    sm_p->ncache_hit = 0;

    ret = PINT_ncache_get_cached_entry(
                (const char*)segment, 
                &object_ref,
                (const PVFS_object_ref*) &parent_ref);

    if (ret == 0)
    {
	struct PVFS_server_resp fake_resp;

	/* build a fake response, call the completion function on it */
	fake_resp.op = PVFS_SERV_LOOKUP_PATH;
	fake_resp.status = 0;
	fake_resp.u.lookup_path.handle_array = &object_ref.handle;
	fake_resp.u.lookup_path.attr_array = NULL;
	fake_resp.u.lookup_path.handle_count = 1;
	fake_resp.u.lookup_path.attr_count = 0;

        gossip_debug(
            GOSSIP_NCACHE_DEBUG, "*** ncache hit on %s (%llu|%d)!\n",
            sm_p->msgpair.req.u.lookup_path.path,
            llu(object_ref.handle), object_ref.fs_id);

        sm_p->ncache_hit = 1;
	ret = sm_p->msgpair.comp_fn(smcb, &fake_resp, 0);
	return 1;
    }

    gossip_debug(GOSSIP_NCACHE_DEBUG, "*** ncache clean miss on %s.\n",
                 sm_p->msgpair.req.u.lookup_path.path);

    js_p->error_code = LOOKUP_NCACHE_MISS;
    return 1;
}

/*
 * Local variables:
 *  mode: c
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
